"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CLASS_CONCEALED = 'concealed';
exports.CLASS_CONCEALING = 'concealing';
exports.CLASS_REVEALING = 'revealing';
var DATASET_ORIGINAL_INLINE_STYLE_HEIGHT = 'concealRevealOriginalInlineStyleHeight';
var DATASET_ORIGINAL_IMPORTANT_INLINE_STYLES = 'concealRevealOriginalImportantInlineStyle';
var MODIFYING_INLINE_STYLE_PROPERTY_NAMES = [
    'overflow-y',
    'min-height',
    'padding-top',
    'padding-bottom',
    'border-top-width',
    'border-bottom-width',
    'margin-top',
    'margin-bottom',
];
/**
 * Returns whether the element is concealed (including transitioning to be concealed).
 */
function concealed(element) {
    return element.classList.contains(exports.CLASS_CONCEALED) || element.classList.contains(exports.CLASS_CONCEALING);
}
exports.concealed = concealed;
/**
 * Conceals the element.
 */
function conceal(element) {
    if (!concealed(element)) {
        var classList = element.classList;
        if (element.offsetHeight) {
            if (element.dataset[DATASET_ORIGINAL_INLINE_STYLE_HEIGHT] === undefined) {
                saveOriginalInlineStyleHeight(element);
            }
            saveOriginalImportantInlineStyles(element);
            // calculate the style height and explicitly set it as inline style
            element.style.setProperty('height', getCurrentStyleHeight(element) + "px", 'important');
            element.offsetHeight; // reflect height above
            // conceal the element with transition
            element.addEventListener('transitionend', onTransitionEnd);
            classList.add(exports.CLASS_CONCEALING);
            classList.remove(exports.CLASS_REVEALING);
            deleteImportantInlineStyle(element);
            element.style.height = '';
        }
        else {
            classList.add(exports.CLASS_CONCEALED);
        }
    }
}
exports.conceal = conceal;
/**
 * Reveals the concealed element.
 */
function reveal(element) {
    if (concealed(element)) {
        var currentOffsetHeight = element.offsetHeight;
        var classList = element.classList;
        element.removeAttribute('aria-hidden');
        // temporarily reveal the element without transition and calculate height
        classList.remove(exports.CLASS_CONCEALING);
        classList.remove(exports.CLASS_CONCEALED);
        if (element.dataset[DATASET_ORIGINAL_INLINE_STYLE_HEIGHT] === undefined) {
            saveOriginalInlineStyleHeight(element);
        }
        else {
            restoreOriginalInlineStyleHeight(element);
        }
        restoreOriginalImportantInlineStyles(element);
        if (element.offsetHeight !== currentOffsetHeight) {
            var expandedStyleHeight = getCurrentStyleHeight(element);
            // conceal the element without transition again
            classList.add(exports.CLASS_CONCEALED);
            deleteImportantInlineStyle(element);
            element.style.height = '';
            element.offsetHeight; // reflect zero-height
            // reveal it with transition
            element.addEventListener('transitionend', onTransitionEnd);
            classList.add(exports.CLASS_REVEALING);
            classList.remove(exports.CLASS_CONCEALED);
            element.style.setProperty('height', expandedStyleHeight + "px", 'important');
            restoreOriginalImportantInlineStyles(element);
        }
        else {
            delete element.dataset[DATASET_ORIGINAL_INLINE_STYLE_HEIGHT];
        }
        delete element.dataset[DATASET_ORIGINAL_IMPORTANT_INLINE_STYLES];
    }
}
exports.reveal = reveal;
/**
 * If the element is concealed, reveals the element, otherwise conceals the element.
 */
function toggle(element) {
    (concealed(element) ? reveal : conceal)(element);
}
exports.toggle = toggle;
function onTransitionEnd() {
    this.removeEventListener('transitionend', onTransitionEnd);
    var classList = this.classList;
    if (classList.contains(exports.CLASS_CONCEALING)) {
        classList.add(exports.CLASS_CONCEALED);
        classList.remove(exports.CLASS_CONCEALING);
        this.setAttribute('aria-hidden', 'true');
    }
    else if (classList.contains(exports.CLASS_REVEALING)) {
        classList.remove(exports.CLASS_REVEALING);
        restoreOriginalInlineStyleHeight(this);
        delete this.dataset[DATASET_ORIGINAL_INLINE_STYLE_HEIGHT];
    }
}
function getCurrentStyleHeight(element) {
    var style = getComputedStyle(element);
    return element.offsetHeight - (style.boxSizing === 'border-box'
        ? 0
        : parseFloat(style.borderTopWidth) +
            parseFloat(style.borderBottomWidth) +
            parseFloat(style.paddingTop) +
            parseFloat(style.paddingBottom));
}
function saveOriginalInlineStyleHeight(element) {
    var priority = element.style.getPropertyPriority('height');
    element.dataset[DATASET_ORIGINAL_INLINE_STYLE_HEIGHT] = "" + (element.style.height || '') + (priority ? ' !' + priority : '');
}
function restoreOriginalInlineStyleHeight(element) {
    var originalHeight = (element.dataset[DATASET_ORIGINAL_INLINE_STYLE_HEIGHT] || '').split(' !');
    element.style.setProperty('height', originalHeight[0], originalHeight[1] || '');
}
function saveOriginalImportantInlineStyles(element) {
    var styleString = '';
    for (var _i = 0, MODIFYING_INLINE_STYLE_PROPERTY_NAMES_1 = MODIFYING_INLINE_STYLE_PROPERTY_NAMES; _i < MODIFYING_INLINE_STYLE_PROPERTY_NAMES_1.length; _i++) {
        var propertyName = MODIFYING_INLINE_STYLE_PROPERTY_NAMES_1[_i];
        if (element.style.getPropertyPriority(propertyName) === 'important') {
            styleString += propertyName + ": " + element.style.getPropertyValue(propertyName) + "; ";
        }
    }
    styleString && (element.dataset[DATASET_ORIGINAL_IMPORTANT_INLINE_STYLES] = styleString.trim());
}
function restoreOriginalImportantInlineStyles(element) {
    var styleString = element.dataset[DATASET_ORIGINAL_IMPORTANT_INLINE_STYLES];
    if (styleString) {
        var pattern = /\s*([^:\s]+)\s*:\s*([^;\s]+)\s*;?/g;
        var result = void 0;
        while (result = pattern.exec(styleString)) {
            element.style.setProperty(result[1], result[2], 'important');
        }
    }
}
function deleteImportantInlineStyle(element) {
    for (var _i = 0, MODIFYING_INLINE_STYLE_PROPERTY_NAMES_2 = MODIFYING_INLINE_STYLE_PROPERTY_NAMES; _i < MODIFYING_INLINE_STYLE_PROPERTY_NAMES_2.length; _i++) {
        var propertyName = MODIFYING_INLINE_STYLE_PROPERTY_NAMES_2[_i];
        if (element.style.getPropertyPriority(propertyName) === 'important') {
            element.style.removeProperty(propertyName);
        }
    }
}
